/**
 * Created with IntelliJ IDEA.
 * @author: jozhua@live.cn
 * @date: 2017/6/2
 * To change this template use File | Settings | File Templates.
 */
import cp from './els';
let hasTouch = ('ontouchstart' in global) || global.DocumentTouch && document instanceof DocumentTouch;

let specialEvents = {}, handlers = [], focusinSupported = 'onfocusin' in global, _zid = 1,
    focus = {focus: 'focusin', blur: 'focusout'},
    hover = {mouseenter: 'mouseover', mouseleave: 'mouseout'},
    eventMethods = {
        preventDefault: 'isDefaultPrevented',
        stopImmediatePropagation: 'isImmediatePropagationStopped',
        stopPropagation: 'isPropagationStopped'
    }, ignoreProperties = /^([A-Z]|returnValue$|layer[XY]$|webkitMovement[XY]$)/, slice = [].slice
    ;
specialEvents.click = specialEvents.mousedown = specialEvents.mouseup = specialEvents.mousemove = 'MouseEvents';
specialEvents.tap = 'Events';
function zid(element) {
    return element._zid || (element._zid = _zid++)
}
function returnFalse() {
    return false
}
function returnTrue() {
    return true
}
function realEvent(type) {
    return hover[type] || (focusinSupported && focus[type]) || type
}
function parse(event) {
    var parts = ('' + event).split('.');
    return {e: parts[0]};
}
function compatible(event, source) {
    if (source || !event.isDefaultPrevented) {
        source || (source = event)

        cp.each(eventMethods, function (name, predicate) {
            var sourceMethod = source[name]
            event[name] = function () {
                this[predicate] = returnTrue;
                return sourceMethod && sourceMethod.apply(source, arguments)
            }
            event[predicate] = returnFalse
        })

        event.timeStamp || (event.timeStamp = Date.now())

        if (source.defaultPrevented !== undefined ? source.defaultPrevented :
                'returnValue' in source ? source.returnValue === false :
                    source.getPreventDefault && source.getPreventDefault())
            event.isDefaultPrevented = returnTrue
    }
    return event
}

function createProxy(event) {
    var key, proxy = {originalEvent: event}
    for (key in event)
        if (!ignoreProperties.test(key) && event[key] !== undefined) proxy[key] = event[key]

    return compatible(proxy, event)
}
function Event(type, props) {
    if (!cp.isString(type)) props = type, type = props.type
    var event = document.createEvent(specialEvents[type] || 'Events'), bubbles = true;
    if (props) bubbles = props && !!props['bubbles'];
    event.initEvent(type, bubbles, true);
    return compatible(event)
}

function addEvent(element, events, fn, data, selector, delegator, capture) {
    var id = zid(element), set = (handlers[id] || (handlers[id] = []));
    events.split(/\s/).forEach(function (event) {
        var handler = parse(event);
        handler.fn = fn;
        handler.sel = selector;
        handler.del = delegator;
        var callback = delegator || fn;
        handler.proxy = function (e) {
            e.data = data;
            var result = callback.apply(element, e._args == undefined ? [e] : [e].concat(e._args));
            if (result === false) e.preventDefault(), e.stopPropagation();
            return result
        };
        handler.i = set.length;
        set.push(handler);
        if ('addEventListener' in element)
            element.addEventListener(realEvent(handler.e), handler.proxy, !!capture)
    })
}

function removeEvent(element, events, fn, capture) {
    (events || '').split(/\s/).forEach(function (event) {
        findHandlers(element, event, fn).forEach(function (handler) {
            if ('removeEventListener' in element) element.removeEventListener(realEvent(handler.e), handler.proxy, !!capture)
        })
    })
}

function findHandlers(element, event, fn, selector) {
    event = parse(event);
    return (handlers[zid(element)] || []).filter(function (handler) {
        return handler
            && (!event.e || handler.e == event.e)
            && (!fn || zid(handler.fn) === zid(fn))
            && (!selector || handler.sel == selector)
    })
}

cp.extend(cp, {
    events: Event,
    support: {
        touch: hasTouch,
        touchEvent: hasTouch ? 'tap' : 'click'
    }
});
cp.extend(cp.fn, {
    on: function (event, selector, data, callback, one) {
        let autoRemove, delegator, $this = this
        if (event && !cp.isString(event)) {
            cp.each(event, function (type, fn) {
                $this.on(type, selector, data, fn, one)
            })
            return $this
        }

        if (!cp.isString(selector) && !cp.isFunction(callback) && callback !== false)
            callback = data, data = selector, selector = undefined;
        if (callback === undefined || data === false)
            callback = data, data = undefined;

        if (callback === false) callback = returnFalse;

        return $this.each(function (_, element) {
            if (one) autoRemove = function (e) {
                removeEvent(element, e.type, callback)
                return callback.apply(this, arguments)
            }

            if (selector) delegator = function (e) {
                var evt, match = cp.el(e.target).closest(selector, element).get(0)
                if (match && match !== element) {
                    evt = cp.extend(createProxy(e), {currentTarget: match, liveFired: element})
                    return (autoRemove || callback).apply(match, [evt].concat(slice.call(arguments, 1)))
                }
            }

            addEvent(element, event, callback, data, selector, delegator || autoRemove)
        })
    }, off: function (event, selector, callback) {
        var $this = this
        if (event && !cp.isString(event)) {
            cp.each(event, function (type, fn) {
                $this.off(type, selector, fn)
            })
            return $this
        }

        if (!cp.isString(selector) && !cp.isFunction(callback) && callback !== false) callback = selector, selector = undefined;
        if (callback === false) callback = returnFalse;
        return $this.each(function () {
            removeEvent(this, event, callback, selector)
        })
    },
    trigger(event, args) {
        event = (cp.isString(event) || cp.isObject(event)) ? cp.events(event) : compatible(event);
        event._args = args
        return this.each(function () {
            // handle focus(), blur() by calling them directly
            if (event.type in focus && typeof this[event.type] == "function") this[event.type]()
            // items in the collection might not be DOM elements
            else if ('dispatchEvent' in this) this.dispatchEvent(event)
            else cp.el(this).triggerHandler(event, args)
        })
    },
    triggerHandler  (event, args) {
        var e, result
        this.each(function (i, element) {
            e = createProxy(cp.isString(event) ? cp.events(event) : event);
            e._args = args
            e.target = element
            cp.each(findHandlers(element, event.type || event), function (i, handler) {
                result = handler.proxy(e);
                if (e.isImmediatePropagationStopped()) return false
            })
        })
        return result
    }
});